//This header includes all classes used
#ifndef CLASSES_H_
#define CLASSES_H_
#include <iostream>
#include <complex>
#include <ctime>
#include "petsc.h"
using namespace std;

#define blockDimension 12

class Parameters
{
public:
	Parameters(PetscInt, PetscInt, PetscInt, double, double, double, double, double, double, int, double, double, int, int, int, int);
	//set (used as argument)
	PetscInt rows;
	PetscInt columns;
	PetscInt idim;
	double energyMin;
	double energyMax;
	double energyStep;
	double thop;
	double vsdMax;
	double vhop;
	double BField;
	int leadSize;
	int mtop;
	int ntop;
	time_t invertStart;
	
	//set (non argument)
	double vsdMin;
	double a;
	int vsdSteps;
	
	//derived
	PetscInt ndiag;
	int numEnergySteps;
	double energy;
	int imax1;
	int imax2;
	int imaxd21;
	int imaxd22;
	int imaxd21i;
	int imaxd22i;
}; //end block

Parameters::Parameters(PetscInt inRows, PetscInt inColumns, PetscInt inIdim, double inEnergyMin, double inEnergyMax, double inEnergyStep, double inThop,
					   double inVsdMin, double inVsdMax, int inVsdSteps, double inVhop, double inBField, int inLeadSize, int inMtop, int inNtop, int inA)
{
	rows = inRows;
	columns = inColumns;
	idim = inIdim;
	energyMin = inEnergyMin;
	energyMax = inEnergyMax;
	energyStep = inEnergyStep;
	thop = inThop;
	vsdMin = inVsdMin;
	vsdMax = inVsdMax;
	vsdSteps = inVsdSteps;
	vhop = inVhop;
	BField = inBField;
	leadSize = inLeadSize;
	mtop = inMtop;
	ntop = inNtop;
	a = inA;
	
	energy = energyMin;
	ndiag = rows/idim;
	numEnergySteps = (energyMax >= 0 ? (energyMax - energyMin) / energyStep + 0.5
			   : (energyMax - energyMin) / energyStep + 1.5);
	imax1 = rows - idim + 1;
	imax2 = rows - idim + 2;
	imaxd21 = rows/2 + 1;
	imaxd22 = rows/2 + 2;
	imaxd21i = rows/2 - idim + 1;
	imaxd22i = rows/2 - idim + 2;	
} //end block.block

//This function acts the same way as the bphs function in the old code.
//The main difference is I combined the bphs function with the earg function as they were only used together.
complex<double> bphs(int a1, int a2, double BField, double atomCoordinates[][3])
{
	double a1x, a1y, a1z, a2x, a2y, a2z;
	double phi1, phi2, ri1, ri2, rhoij;
	double cearg, searg, earg = 0;
	complex<double> result;
	
    a1x = atomCoordinates[a1-1][0];
    a1y = atomCoordinates[a1-1][1];
    a1z = atomCoordinates[a1-1][2];
    a2x = atomCoordinates[a2-1][0];
    a2y = atomCoordinates[a2-1][1];
    a2z = atomCoordinates[a2-1][2];
	
    rhoij = (a1x + a2x)*(a1x + a2x) + (a1y + a2y)*(a1y + a2y);
	rhoij = .5 * sqrt(rhoij);
	
	if(rhoij < 0.00001)
		earg = 0.0;
	else
	{
		phi1 = -(a1y + a2y) /rhoij/2.0;
		phi2 = (a1x + a2x) /rhoij/2.0;
		ri1 = a1x - a2x;
		ri2 = a1y - a2y;
		earg = phi1*ri1 + phi2*ri2;
	} //endelse
	
	earg = .5 * BField * rhoij * earg;
	earg = 1.52e-5 * earg;
	searg = sin(earg);
	cearg = cos(earg);
	result = complex<double>(cearg, searg);
	
	return result;
} //end bphs


class block
{
	int mi, id;
	complex<double> data[blockDimension][blockDimension];
	char type;
  public:
	void set_values (const PetscInt, const PetscInt, const PetscInt, double, int, int, double, double, char, 
					 double atomCoordinates[][3]);
	void set_data (int, int, complex<double>);
	complex<double> getData(int i, int j){return data[i][j];}
	char getType(){return type;}
	int area () {return (mi*id);}
	void print_block(const PetscInt);
}; //end block


//This is a simple function. Takes the value C and places at a,b in data. Note that a and b much be less than 12
void block::set_data(int a, int b, complex<double> c)
{
	data[a][b] = c;
}  //end block.set_data


//Prints the 12 by 12 data block to the screen.
void block::print_block(const PetscInt idim)
{
	for(int i=0; i<idim; i++)
	{
		for(int j=0; j<idim; j++)
		{
			cout << data[i][j];
			cout << endl;
		} //endfor
	} //endfor
} //end block.print_block


//Sets not only the data, but also the id, and mi. Furthermore, it sets the data to the values determined by its type.
//For instance, the if type = "D" then set_values sets the data along the main diagonal, the two sub-diagonals, and the two corner elements.
void block::set_values(const PetscInt rows, const PetscInt idim, const PetscInt ndiag, double vhop, 
					   int a, int b, double c, double BField, char d, double atomCoordinates[][3])
{
	mi = a;
	type = d;
	bool check;
	const PetscInt itor = idim/2;
    int ic[itor+1];
	id = mi + b*ndiag;
	
	complex<double> diaener;
	diaener = complex<double>(c, 5e-9);
	
	//Reset all data in block to zero, over riding any previous data.
	for(int j=0; j<idim; j++)
	{
		for(int i=0; i<idim; i++)
		{
			data[i][j] = 0.0;
		} //endfor
	} //endfor

	switch(type)
	{
		case 'D' :
			//Setting the diagonal
			for(int i=0; i<idim; i++)
			{
				data[i][i] = diaener;
			} //endfor

			for(int i=1; i<idim; i++)
			{
				//Above diagonal
				data[i-1][i] = bphs(idim*mi+i-1+1, idim*mi+i+1, BField, atomCoordinates) * (-vhop);
				//Below diagonal
				data[i][i-1] = bphs(idim*mi+i+1, idim*mi+i-1+1, BField, atomCoordinates) * (-vhop);
			} //endfor

			data[0][idim-1] = -vhop;
			data[idim-1][0] = -vhop;
			break;

		case 'W' :
			//Setting the t-dagger blocks.
			check = true;
			for(int j=1; j<idim; j+=2)
			{
				if(check == true)
				{
					data[j-1][j] = bphs(mi*idim+j+idim, mi*idim+j+1, BField, atomCoordinates) * (-vhop); //1,5,9
					check = !check;
				} //endif
				else if(check == false)
				{
					data[j][j-1] = bphs(mi*idim+j+idim+1, mi*idim+j, BField, atomCoordinates) * (-vhop); //3,7,11
					check = !check;
				} //endif
			} //endfor
			break;

		case 'T' :
			//Setting the T blocks
			check = true;
			for(int j=1; j<idim; j+=2)
			{
				if(check == true)
				{
					data[j][j-1] = conj(bphs(mi*idim+j+idim, mi*idim+j+1, BField, atomCoordinates)) * (-vhop); //3,7,11
					check = !check;
				} //endif
				else if(check == false)
				{
					data[j-1][j] = conj(bphs(mi*idim+j+idim+1, mi*idim+j, BField, atomCoordinates)) * (-vhop); //1,5,9
					check = !check;
				} //endif
			} //endfor
			break;

		case 'U' :
			//Setting the U blocks
			check = true;
			for(int j=1; j<=itor; j++)
			{
				if(check == true)
				{
					ic[j] = 2*j-1;
					data[ic[j]-1][ic[j]] = bphs(2*j-1, rows-idim+2*(j-1)+2, BField, atomCoordinates) * (-vhop);
					check = !check;
				} //endif
				else if(check == false)
				{
					ic[j] = 2*(j-1);
					data[ic[j]+1][ic[j]] = bphs(2*j, rows-idim+2*(j-1)+1, BField, atomCoordinates) * (-vhop);
					check = !check;
				} //endif
			} //endfor
			//data[0][1] = bphs(1, 3590, BField, atomCoordinates) * (-vhop);
			//data[3][2] = bphs(4, 3591, BField, atomCoordinates) * (-vhop);
			//data[4][5] = bphs(5, 3594, BField, atomCoordinates) * (-vhop);
			//data[7][6] = bphs(8, 3595, BField, atomCoordinates) * (-vhop);
			//data[8][9] = bphs(9, 3598, BField, atomCoordinates) * (-vhop);
			//data[11][10] = bphs(idim, 3599, BField, atomCoordinates) * (-vhop);
			break;

		case 'Q' :
			//Setting the u dagger blocks
			check = true;
			for(int j=1; j<=itor; j++)
			{
				if(check == true)
				{
					ic[j] = 2*j-1;
					data[ic[j]][ic[j]-1] = bphs(rows-idim+2*(j-1)+2, 2*j-1, BField, atomCoordinates) * (-vhop);
					check = !check;
				} //endif
				else if(check == false)
				{
					ic[j] = 2*(j-1);
					data[ic[j]][ic[j]+1] = bphs(rows-idim+2*(j-1)+1, 2*j, BField, atomCoordinates) * (-vhop);
				check = !check;
				} //endif
			} //endfor
			//data[1][0] = bphs(3590, 1, BField, atomCoordinates) * (-vhop);
			//data[2][3] = bphs(3591, 4, BField, atomCoordinates) * (-vhop);
			//data[5][4] = bphs(3594, 5, BField, atomCoordinates) * (-vhop);
			//data[6][7] = bphs(3595, 8, BField, atomCoordinates) * (-vhop);
			//data[9][8] = bphs(3598, 9, BField, atomCoordinates) * (-vhop);
			//data[10][11] = bphs(3599, idim, BField, atomCoordinates) * (-vhop);
			break;
	}
} //end block.set_values

#endif /* CLASSES_H_ */
